
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   rec_bisect.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Mon Jun 14 21:58:16 2010
 * 
 * @brief  1d recursive bisection
 * 
 * 
 */


#ifndef __REC_BISECT_H__
#define __REC_BISECT_H__
#include <util/prefix_sum_tools.hpp>
#include <assert.h>
#include <stdlib.h>
#include <util/tools.hpp>
#include <iostream>
#include <stdio.h>


namespace oned
{
/**
 * @brief recursive 1D partitioning algorithm which partition an array
 * in two parts of similiar load and allocates half the processor in
 * each part.
 *
 * @param T data type of the Prefix Sum Array
 * @param Pr type of the Prefix Sum Array
 */

template <typename T, typename Pr>
class RecursiveBisection
{
  static const bool debugFEC=false;
  static const bool debugRB=false;


  /*Returns the recursive bisection cuts in sorted order in
   *cutPoints. low and high included.*/
  static T rec_bisection_internal(int numProc, int low, int high,
				  const Pr& prefixSum, int *cutPoints);
  

public:
  /*Calculates the abstract weight difference of two partitions which
    are seperated by the index "candidate". candidate remains in the
    left processor. prefixsum[i] is the sum of first i elements in a
    given array. prefixsum[0] = 0. low and high values are included
    while calculating the difference. rightProc is the number of
    processors on the right and leftProc is the number of processors on
    the left*/
  static double calculateDiff(const Pr& prefixSum, int low, int high,
			      int leftProc, int rightProc, int candidate);




  /** 
   * Finds the best cut point of a given array by calculating  Wtot*(leftProc)/(leftProc+rightProc) and then decides whether to cut from left or right. left and low index values are included to calculation. prefixsum[0] = 0.
   * 
   * @param prefixSum 1D prefixsum array
   * @param low lower bound
   * @param high upper bound
   * @param leftProc number of processor on left side
   * @param rightProc number of processor on right side
   * 
   * @return cut index (corresponds to the last element of leftProc and one earlier than first element of rightProc) 
   */
  static int findEvenCut(const Pr& prefixSum, int low, int high,
			 int leftProc, int rightProc);  
  /** 
   * 
   * 
   * @param[in] procCount number of processors 
   * @param[in] prefixSumArray Prefix sum array that always starts with value 0 in the 0th index
   * @param[in] length is the length of prefixSumArray 
   * @param[out] cutIndexes cut index points
   * @param[in] max maximum load of a given array. If unknown, use -1. (ignored in this function.)
   * 
   * @return Load of the max. loaded processor
   */
  static T rec_bisection(int procCount, const Pr& prefixSumArray,
			 int length, int *cutIndexes, T max);
};
}
#include <oned/rec_bisect_impl.hpp>

#endif
