
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   rec_bisect.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Mon Jun 14 21:58:16 2010
 * 
 * @brief  1d recursive bisection
 * 
 * 
 */


#ifndef __REC_BISECT_MINMAX_H__
#define __REC_BISECT_MINMAX_H__
#include <util/prefix_sum_tools.hpp>
#include <assert.h>
#include <cstdlib>
#include <util/tools.hpp>
namespace oned
{
/**
 * @brief 1D relaxed hierarchical heuristic.
 *
 * @param T data type of the Prefix Sum Array
 * @param Pr type of the Prefix Sum Array
 */
template <typename T, typename Pr>
class RecursiveBisectionMinMax
{
  //For debug purpose only
  static int linsrch(const Pr prefixSum, int low, int high, int leftProc, int rightProc);

  static double pointDiff(const Pr prefixSum, int low, int high, int leftProc, int rightProc, int point);
/**
 * @brief Finds the slope sign at a given point
 * @param prefixSum 1D prefix sum array
 * @param low Lowest index limit of prefix sub-array
 * @param high Highest index limit of prefix sub-array
 * @param leftProc number of procs to be assigned to the left
 * @param rightProc number of procs to be assigned to the right 
 * @param point The index of cut point that we want to retrieve Bottleneck/Cut point sign
 * 
 * @return
 * Returns a signature
 * -1 means we are in decreasing side of the function
 * +1 means we are in increasing side of the function 
 */
  static int slopeAt(const Pr prefixSum, int low, int high, int leftProc, int rightProc, int point);
  static int bimonsearch(const Pr prefixSum, int low, int high, int leftProc, int rightProc);
  /*Returns the recursive bisection cuts in sorted order in
   *cutPoints. low and high included.*/
  static T rec_bisection_internal(int numProc, int low, int high,
				  const Pr prefixSum, int *cutPoints);
  

public:
  /*Calculates the abstract weight difference of two partitions which
    are seperated by the index "candidate". candidate remains in the
    left processor. prefixsum[i] is the sum of first i elements in a
    given array. prefixsum[0] = 0. low and high values are included
    while calculating the difference. rightProc is the number of
    processors on the right and leftProc is the number of processors on
    the left*/
  static double calculateDiff(const Pr prefixSum, int low, int high,
			      int leftProc, int rightProc, int candidate);




  /** 
   * Finds the best cut point of a given array by calculating  Wtot*(leftProc)/(leftProc+rightProc) and then decides whether to cut from left or right. left and low index values are included to calculation. prefixsum[0] = 0.
   * 
   * @param[in] prefixSum 1D prefixsum array
   * @param[in] low lower bound
   * @param[in] high upper bound
   * @param[in] leftProc number of processor on left side
   * @param[in] rightProc number of processor on right side
   * 
   * @return index of even cut point (the index of leftProc's last element)
   */
  static int findEvenCut(const Pr prefixSum, int low, int high,
			 int leftProc, int rightProc);  
  /** 
   * Performs 1D recursive bisection algorithm
   * 
   * @param[in] procCount total number of processors
   * @param[in] prefixSumArray prefixSumArray
   * @param[in] length length of prefixSumArray
   * @param[out] cutIndexes cut point results
   * 
   * @return Load of the max. loaded processor
   */
  static T rec_bisection(int procCount, const Pr prefixSumArray,
			 int length, int *cutIndexes, T);
};
}
#include <oned/rec_bisect_minmax_impl.hpp>

#endif
