
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   her_jag_2d.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Mon Jun 14 06:11:04 2010
 * 
 * @brief  2d Heuristic Jagged paritioning algorithm
 * 
 * 
 */


#ifndef __HER_JAG_2D_H__
#define __HER_JAG_2D_H__

#include <util/prefix_sum_tools.hpp>
#include <util/rect_list.hpp>
#include <twod/part_base.hpp>
#include <oned/nicol_plus.hpp>
#include <util/transpose2d.hpp>

namespace twod
{
  /**
   * @brief implements a PxQ jagged partitioning heuristic using the
   * first dimension as main dimension.
   *
   * The heuristic partitions the main dimension in a number of
   * stripes given by setP(). The elements along the auxilary
   * dimension are (virtually) summed into a one dimensionnal array
   * which is partitioned using onedalgoX (oned::NicolPlus by
   * default). Then each stripe is partitioned using onedalgoY
   * (oned::NicolPlus by default).
   *
   * This heurisitc has been proposed numerous times for instance in
   * "Ali Pinar and Cevdet Aykanat, Sparse Matrix Decomposition with
   * Optimal Load Balancing,  HiPC 1997"
   *
   * @param T data type of instance matrix
   * @param Pr data type of 2D matrix
   * @param *onedalgoY algorithm to be used to fix columns
   * @param *onedalgoX algorithm to be used to fix rows
   */
  template <typename T, typename Pr, 
	    T (*onedalgoY) (int procCount, const util::Aggreg2Dto1D<T, Pr, false>& prefixSumArray, int length, int *cutIndexes, T) = oned::NicolPlus<T, util::Aggreg2Dto1D<T, Pr, false> >::nicol_plus,
	    T (*onedalgoX) (int procCount, const util::Aggreg2Dto1D<T, Pr, true>& prefixSumArray, int length, int *cutIndexes, T) = oned::NicolPlus<T, util::Aggreg2Dto1D<T, Pr,  true> >::nicol_plus >
  class JagPQHeurHor : public PartBase<T,Pr>
  {
    int P; //number of stripes in the main dimension. 0 stands for automatic.
  public: 
    JagPQHeurHor();
    static T her_jag_2d_internal(int procX, int procY, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);

    /** 
     * 
     * Applies 2d heuristic jagged partitioning to a given 2d prefixSumArray 
     * @param[in] procCount is the number of processors
     * @param[in] prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param[out] parts  must be allocated before calling this function. List of rectangles that form the partition.
     * 
     * @return Total load of the maximum loaded rectangle.
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);

    /**
     * @brief selects the number of stripes in the main dimension
     *
     * @param P number of stripes in the main dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. 
     */
    void setP(int P);
    
    virtual ~JagPQHeurHor();
  };



  
  /**
   * @brief implements a PxQ jagged partitioning heuristic using the
   * second dimension as main dimension.
   *
   * This class is implemented as a wrapper to twod::JagPQHeurHor
   * using util::TransposePrefix2D.
   *
   * @param T data type of instance matrix
   * @param Pr data type of 2D matrix
   */
  template <typename T, typename Pr,
	    T (*onedalgoY) (int procCount, const util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, false>& prefixSumArray, int length, int *cutIndexes, T) = oned::NicolPlus<T, util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, false> >::nicol_plus,
	    T (*onedalgoX) (int procCount, const util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, true>& prefixSumArray, int length, int *cutIndexes, T) = oned::NicolPlus<T, util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>,  true> >::nicol_plus 
	    >
  class JagPQHeurVer : public PartBase<T,Pr>
  {
    int P;
  public: 
    JagPQHeurVer();
    virtual ~JagPQHeurVer();
    
    /**
     * @brief selects the number of stripes in the main dimension
     *
     * @param P number of stripes in the main dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. 
     */
    void setP(int P);
    /** 
     * 
     * Applies 2d heuristic jagged partitioning to a given 2d
     * prefixSumArray
     *
     * @param procCount is the number of processors
     * @param prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param parts  must be allocated before calling this function
     * 
     * @return *parts, a rectangle list of partitions. Borders of the rectangle is incuded when calculating total load of a rectangle
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);  
  };



  /**
   * @brief implements a PxQ jagged partitioning heuristic taking as
   * the main dimension the one that leads to the best load balance.
   *
   * This class is implemented as a wrapper to twod::JagPQHeurHor
   * and twod::JagPQHeurVer.
   *
   * @param T data type of instance matrix
   * @param Pr data type of 2D matrix
   */
  template <typename T, typename Pr>
  class JagPQHeurBest : public PartBase<T,Pr>
  {
  public: 
    
    virtual ~JagPQHeurBest();
    /** 
     * 
     * @brief Applies 2d heuristic jagged partitioning to a given 2d prefixSumArray 
     *
     * @param[in] procCount is the number of processors
     * @param[in] prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param[out] parts  must be allocated before calling this function. List of rectangles that form the partition.
     * 
     * @return Total load of the maximum loaded rectangle.
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);  
  };

}

#include <twod/her_jag_2d_impl.hpp>

#endif
