
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

#ifndef __TWOD__HYBRID__
#define __TWOD__HYBRID__

#include <util/subpsa.hpp>
#include <util/constraint.hpp>
#include <util/proportional.hpp>
#include <util/rect_list.hpp>
#include <twod/part_base.hpp>
#include <math.h>
#include <memory>

namespace twod
{
  /**
   * @brief hybrid algorithms for 2d partitioning. A first
   * partitioning is done with a 2d algorithm. Then, each part is
   * partitioned during a second level using an other algorithm.
   *
   * At first level, the number of processors used at the first level
   * is variable (but can be fixed manually using setProcLevel1()). At
   * second level, the processors are distributed among the parts
   * proportionnally to their load (using
   * util::distribute_processors).
   *
   * The algorithm at level 1 is selected using setLevel1().
   *
   * To obtain better quality tradeoff, one can use two algorithms at
   * the second phase. Select the fast algorithm using setLevel2() and
   * the slow algorithm using setLevel2Slow(). The slow algorithm will
   * only be used if the fast algorithm does not gets a good load
   * balance.
   *
   * @param T type of the load
   * @param Pr type of the prefix sum array
   */
  template<typename T, typename Pr>
  class Hybrid : public PartBase<T,Pr>
  {
    typedef util::SubPSA<T, Pr> sPr;

    std::auto_ptr<twod::PartBase<T,Pr> > algoLevel1;
    int s1;
    std::auto_ptr<twod::PartBase<T, sPr > > algoLevel2;
    std::auto_ptr<twod::PartBase<T, sPr > > algoLevel2Slow;

  private:
    void probephase1(int procCount,const Pr & prefixSumArray, util::RectList<T, Pr > & l1);
    ///this function returns the maximum p' such that ceil(m-p')/p' = ceil(m-p)/p
    int better_guess(int p, int m);

    void phase1(int procCount,const Pr & prefixSumArray, util::RectList<T, Pr > & l1);

    void phase2(int procCount,const Pr & prefixSumArray,
		util::RectList< T, Pr > & parts, const util::RectList<T, Pr > & l1);


  public:
    /**
     * @brief set the algorithm at the first level.
     *
     * Take care, the algorithm is passed through a pointer to an
     * object. Hybrid WILL deallocate it (through a std::auto_ptr).
     */
    void setLevel1(twod::PartBase<T,Pr> * l1);

    /**
     * @brief set the number of parts to generate at the first level.
     *
     * If -1 is selected, the number of parts at the first level is
     * automatically selected by running the first level multiple
     * times and computing an expected gain at level 2.
     *
     * Note: -1 is the default value.
     */
    void setProcLevel1(int s1);
    
    /**
     * @brief set the algorithm at the second level.
     *
     * Take care, the algorithm is passed through a pointer to an
     * object. Hybrid WILL deallocate it (through a std::auto_ptr).
     */
    void setLevel2(twod::PartBase<T, sPr  > * l2);

    /**
     * @brief set the algorithm at the second level for the difficult
     * case.
     *
     * Take care, the algorithm is passed through a pointer to an
     * object. Hybrid WILL deallocate it (through a std::auto_ptr).
     */
    void setLevel2Slow(twod::PartBase<T, sPr  > * l2);
    /** 
     * 
     * Applies 2 level hybrid partitioning
     * @param[in] procCount is the number of processors
     * @param[in] prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param[out] parts  must be allocated before calling this function. List of rectangles that form the partition.
     * 
     * @return Total load of the maximum loaded rectangle.
     */
    virtual T part (int procCount, const Pr & prefixSumArray, util::RectList< T, Pr > & parts);

    virtual ~Hybrid();

  };
}
#include <twod/hybrid_impl.hpp>
#endif
