
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

#ifndef __JAG_PQ_OPT_INTERVAL__
#define __JAG_PQ_OPT_INTERVAL__

#include <util/aggreg2dto1dpart.hpp>
#include <twod/part_base.hpp>

namespace twod
{

  /**
   * @brief optimal PxQ jagged algorithm from "Ali Pinar and Cevdet
   * Aykanat, Sparse Matrix Decomposition with Optimal Load Balancing,
   * HiPC 1997" using the first dimension as the main dimension.
   *
   * It uses oned::Nicol_plus_interval to partition the first
   * dimension using an optimal partitioning algorithm in the second
   * dimension. That way it gets an accurate value for the load of an
   * interval.
   *
   * The number of stripes in the main dimension can be changed using
   * setP().
   *
   * @param T type of the data in the prefix sum array
   * @param Pr type of the prefixsum array
   */
  template <typename T, typename Pr>
  class JagPQOptIntervalHor : public PartBase<T, Pr>
  {
    int P;

    static T internal(int procX, int procY, const Pr& prefixSumArray, util::RectList<T, Pr > &parts);

  public:

    virtual ~JagPQOptIntervalHor();
    JagPQOptIntervalHor();

    /**
     * @brief selects the number of stripes in the main dimension
     *
     * @param P number of stripes in the main dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. 
     */
    void setP(int P);
    /** 
     *   
     * @brief optimal PxQ jagged algorithm from "Ali Pinar and Cevdet
     * Aykanat, Sparse Matrix Decomposition with Optimal Load
     * Balancing, HiPC 1997" using the first dimension as the main
     * dimension.
     *
     * It uses oned::Nicol_plus_interval to partition the first
     * dimension using an optimal partitioning algorithm in the second
     * dimension. That way it gets an accurate value for the load of an
     * interval.
     * 
     * @param[in] procCount number of processors in the partition.
     * @param[in] prefixSumArray first column and first row consists of
     * zero only. But the borders of rectangles in rect_list never
     * touch this area (index 0 in row or column)
     * @param[out] parts must be allocated before calling this function.
     * 
     * @return the load of maximum loaded processor.
     */
    virtual T part (int procCount, const Pr& prefixSumArray, util::RectList<T, Pr > &parts);
  };    
  
  /**
   * @brief optimal PxQ jagged algorithm from "Ali Pinar and Cevdet
   * Aykanat, Sparse Matrix Decomposition with Optimal Load Balancing,
   * HiPC 1997" using the second dimension as the main dimension.
   *
   * This class is implemented as a wrapper to
   * twod::JagPQOptIntervalHor.
   *
   * @param T type of the data in the prefix sum array
   * @param Pr type of the prefixsum array
   */
  template<typename T, typename Pr>
  class JagPQOptIntervalVer : public PartBase<T,Pr>
  {
    int P;
  public:  
    JagPQOptIntervalVer();


    /**
     * @brief selects the number of stripes in the main dimension
     *
     * @param P number of stripes in the main dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. 
     */
    void setP(int P);

    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    
    virtual ~JagPQOptIntervalVer();
  };
  
  /**
   * @brief optimal PxQ jagged algorithm from "Ali Pinar and Cevdet
   * Aykanat, Sparse Matrix Decomposition with Optimal Load Balancing,
   * HiPC 1997" . It selects the main dimension to get the best
   * possible load balance.
   *
   * This class is implemented as a wrapper to
   * twod::JagPQOptIntervalHor and twod::jag_pq_opt_nicol_ver.
   *
   * @param T type of the data in the prefix sum array
   * @param Pr type of the prefixsum array
   */
  template<typename T, typename Pr>
  class JagPQOptIntervalBest : public PartBase<T,Pr>
  {
  public:  
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    
    virtual ~JagPQOptIntervalBest();
  };

}

#include <twod/jag_pq_opt_interval_impl.hpp>
#endif
