
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

#ifndef __JAGGED_DP_2D_H__
#define __JAGGED_DP_2D_H__

#include <util/prefix_sum_tools.hpp>
#include <util/tools.hpp>
namespace twod
{
  /**
   * @brief An optimal m-way jagged algorithm based on dynamic
   * programming.
   * 
   * It takes the first dimension as the main dimension.  Despite it
   * is well optimized it is still fairly slow. It is based on the
   * following dynamic programming formulation: \f$f(n,m) = \min_{1
   * \leq x \leq n} \min_{0 \leq p \leq m-1} \max (f (x-1, p),
   * algo\_1d ([x:n], m-p))\f$.
   *
   * @param T type of the load
   * @param Pr type of the prefix sum array
   */
  template<typename T, typename Pr>
  class JagMOptHor  : public PartBase<T,Pr>
  {
  private:
    static T jagged_dp_internal(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    
  public:  
    JagMOptHor(){}

    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    
    virtual ~JagMOptHor(){}
  };
  
  /**
   * @brief An optimal m-way jagged algorithm based on dynamic
   * programming.
   * 
   * It takes the second dimension as the main dimension.
   * This class is implemented as a wrapper to twod::JagMOptHor.
   *
   * @param T type of the load
   * @param Pr type of the prefix sum array
   */
  template<typename T, typename Pr>
  class JagMOptVer  : public PartBase<T,Pr>
  {
  public:  
    /** 
     * 
     * @brief Applies 2D jagged-optimal partitioning
     * @param[in] procCount is the number of processors
     * @param[in] prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param[out] parts  must be allocated before calling this function. List of rectangles that form the partition.
     * 
     * @return Total load of the maximum loaded rectangle.
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    
    virtual ~JagMOptVer(){}
  };
  
  /**
   * @brief An optimal m-way jagged algorithm based on dynamic
   * programming.
   * 
   * It selects the main dimension to get the best load.
   * This class is implemented as a wrapper to twod::JagMOptHor
   * and twod::JagMOptVer.
   *
   * @param T type of the load
   * @param Pr type of the prefix sum array
   */
  template<typename T, typename Pr>
  class JagMOptBest  : public PartBase<T,Pr>
  {
  public:  
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    
    virtual ~JagMOptBest(){}
  };
  

}
#include <twod/jagged_dp_impl.hpp>

#endif
