
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

#ifndef __JAGGED_DP_2D_PQ_H__
#define __JAGGED_DP_2D_PQ_H__

#include <util/prefix_sum_tools.hpp>
#include <twod/part_base.hpp>

namespace twod
{
  /**
   * @brief an optimal PxQ-way jagged partitionning algorithm using a
   * Dynamic Programming based algorithm.
   *
   * It is inspired from the algorithm in "Fredrik Manne and Tor
   * Sorevik, Partitioning an Array onto a Mesh of Processors, PARA
   * 1996". It takes the first dimension as the main dimension.
   *
   * @param T type of the load
   * @param Pr type of the prefix sum array
   */
  template<typename T, typename Pr>
  class JagPQOptHor : public PartBase<T,Pr>
  {
  private:
    static T jagged_dp_pq_internal(int P, int Q, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);

    int P;//select number of cut in the main dimension. 0 for automatic
  public:  
    JagPQOptHor():P(0){}

    /**
     * @brief selects the number of stripes in the main dimension.
     *
     * @param P number of stripes in the first dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. 
     */
    void setP(int P){this->P=P;}
    /** 
     * 
     * Applies 2D jagged-optimal partitioning. The partition ends up with P rows that has Q processor assigned to each row.
     * @param[in] procCount is the number of processors
     * @param[in] prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param[out] parts  must be allocated before calling this function. List of rectangles that form the partition.
     * 
     * @return Total load of the maximum loaded rectangle.
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);

    virtual ~JagPQOptHor(){}
  };

  /**
   * @brief an optimal PxQ-way jagged partitionning algorithm using a
   * Dynamic Programming based algorithm. 
   *
   * It is inspired from the algorithm in "Fredrik Manne and Tor
   * Sorevik, Partitioning an Array onto a Mesh of Processors, PARA
   * 1996". It takes the second dimension as the main dimension.
   * This classes implemets the algorithm as a wrapper to twod::JagPQOptHor.
   *
   * @param T type of the load
   * @param Pr type of the prefix sum array
   */
  template<typename T, typename Pr>
  class JagPQOptVer  : public PartBase<T,Pr>
  {
    int P;
  public:  
    JagPQOptVer():P(0){}

    /**
     * @brief selects the number of stripes in the main dimension.
     *
     * @param P number of stripes in the first dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. 
     */
    void setP(int P){this->P=P;}
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    
    virtual ~JagPQOptVer(){}
  };
  
  /**
   * @brief an optimal PxQ-way jagged partitionning algorithm using a
   * Dynamic Programming based algorithm.
   *
   * It is inspired from the algorithm in "Fredrik Manne and Tor
   * Sorevik, Partitioning an Array onto a Mesh of Processors, PARA
   * 1996". It selects the main dimension to get the best load
   * balance.  This classes implemets the algorithm as a wrapper to
   * twod::JagPQOptHor and twod::JagPQOptVer.
   *
   * @param T type of the load
   * @param Pr type of the prefix sum array
   */
  template<typename T, typename Pr>
  class JagPQOptBest  : public PartBase<T,Pr>
  {
  public:  
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    
    virtual ~JagPQOptBest(){}
  };

}
#include <twod/jagged_dp_pq_impl.hpp>

#endif
