
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   m_way_jag_2d.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Mon Jun 26 11:11:04 2010
 * 
 * @brief  M-way Jagged paritioning algorithm
 * 
 * 
 */


#ifndef __M_WAY_JAG_2D_H__
#define __M_WAY_JAG_2D_H__

#include <util/prefix_sum_tools.hpp>
#include <util/rect_list.hpp>
#include <util/transpose2d.hpp>
#include <oned/nicol_plus.hpp>

namespace twod
{
  

  /**
   * @brief implements an m-way jagged partitioning heuristic using
   * the first dimension as the main dimension.
   *
   * The heuristic partitions the first dimension in a number of
   * stripes given by setP() using onedalgoX (default:
   * oned::NicolPlus). Then it distributes the processors to the
   * stripes proportionnally to their load. Finally, each stripe is
   * partitioned using onedalgoY (default oned::NicolPlus).
   *
   * @param T data type of instance matrix
   * @param Pr data type of 2D matrix
   * @param onedalgoY 1d partitioning algorithm to be used to fix columns. If unspecified, uses oned::NicolPlus
   * @param onedalgoX 1d partitioning algorithm to be used to fix rows. If unspecified, uses oned::NicolPlus
   */
  template <typename T, typename Pr, 
	    T (*onedalgoY) (int procCount, const util::Aggreg2Dto1D<T, Pr, false>& prefixSumArray, int length, int *cutIndexes, T)=oned::NicolPlus<T, util::Aggreg2Dto1D<T, Pr, false> >::nicol_plus ,
	    T (*onedalgoX)(int procCount, const util::Aggreg2Dto1D<T, Pr, true>& prefixSumArray, int length, int *cutIndexes, T)=oned::NicolPlus<T, util::Aggreg2Dto1D<T, Pr, true> >::nicol_plus >
  class JagMHeurHor : public PartBase<T,Pr>
  {
  
    int P;//chose the number of stripe. 0 is automatic.
  public: 
    JagMHeurHor():P(-1){}

    /**
     * @brief selects the number of stripes in the main dimension.
     *
     * @param P number of stripes in the main dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. If -1, the number of
     * stripes is ceil(sqrt(m*prefixsizeX/prefixsizeY)). -1 is the
     * default.
     */
    void setP(int P){this->P=P;}
    /** 
     * 
     * @brief Applies m-way jagged partitioning to a given 2d prefixSumArray 
     * @param[in] procCount is the number of processors
     * @param[in] prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param[out] parts  must be allocated before calling this function
     * 
     * @return The load of maximum loaded processor.
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    virtual ~JagMHeurHor(){}


    /**
     * @brief actual implementation of the algorithm.
     *
     * @param procX number of stripes
     * @param m total number of processors
     */
    static T m_way_jag_2d_internal(int procX, int m, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
  };


  /**
   * @brief implements an m-way jagged partitioning heuristic using
   * the second dimension as the main dimension.
   *
   * This class is implemented as a wrapper to twod::JagMHeurHor.
   *
   * @param T data type of instance matrix
   * @param Pr data type of 2D matrix
   * @param onedalgoY 1d partitioning algorithm to be used to fix columns. If unspecified, uses oned::NicolPlus
   * @param onedalgoX 1d partitioning algorithm to be used to fix rows. If unspecified, uses oned::NicolPlus
   */
  template <typename T, typename Pr,
	    T (*onedalgoY) (int procCount, const util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, false>& prefixSumArray, int length, int *cutIndexes, T) = oned::NicolPlus<T, util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, false> >::nicol_plus,
	    T (*onedalgoX) (int procCount, const util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, true>& prefixSumArray, int length, int *cutIndexes, T) = oned::NicolPlus<T, util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>,  true> >::nicol_plus 
	    >
  class JagMHeurVer : public PartBase<T,Pr>
  {
    int P;
  public: 
    JagMHeurVer():P(0){}
    virtual ~JagMHeurVer(){}
    
    /**
     * @brief selects the number of stripes in the main dimension.
     *
     * @param P number of stripes in the main dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. If -1, the number of
     * stripes is ceil(sqrt(m*prefixsizeX/prefixsizeY)). -1 is the
     * default.
     */
    void setP(int P)
    {
      this->P = P;
    }
    /** 
     * 
     * Applies 2d heuristic jagged partitioning to a given 2d prefixSumArray 
     * @param[in] procCount is the number of processors
     * @param[in] prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param[out] parts  must be allocated before calling this function
     * 
     * @return the load of the most loaded rectangle in the generated partition.
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);  
  };



  /**
   * @brief implements an m-way jagged partitioning heuristic using
   * as the main dimension the one that leads to the best load balance.
   *
   * This class is implemented as a wrapper to twod::JagMHeurHor and twod::JagMHeurVer.
   *
   * @param T data type of instance matrix
   * @param Pr data type of 2D matrix
   */
  template <typename T, typename Pr>
  class JagMHeurBest : public PartBase<T,Pr>
  {
  public: 
    
    virtual ~JagMHeurBest(){}
    /** 
     * Applies 2d heuristic jagged partitioning to a given 2d prefixSumArray 
     * @param[in] procCount is the number of processors
     * @param[in] prefixSumArray first column and first row consists of zero only. But the borders of rectangles in rect_list never touch this area (index 0 in row or column)
     * @param[out] parts  must be allocated before calling this function
     * 
     * @return the load of the most loaded rectangle in the generated partition.
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);  
  };
  
}


#include <twod/m_way_jag_2d_impl.hpp>

#endif



