
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   m_way_probe.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Sat Jun 11 04:07:10 2011
 * 
 */


#ifndef __TWOD_M_WAY_PROBE__
#define __TWOD_M_WAY_PROBE__

#include <util/m_probe.hpp>
#include <util/prefix_sum_tools.hpp>
#include <twod/part_base.hpp>
#include <vector>
#include <util/multiarrays.hpp>
#include <util/transpose2d.hpp>

namespace twod
{


  
  /**
   * @brief Finds least max proc load value by applying binary search
   * over it. First cut direction is horizontal. See section 3.2.2
   * PROBE-M algorithm of  <a href="http://arxiv.org/abs/1104.2566"> "E. Saule,
   * E. O. Bas, U. V. Catalyurek, Load-Balancing Spatially Located
   * Computations using Rectangular Partitions Paper, Tech. Report,
   * ArXiv, no. arXiv:1104.2566v1, 2011"</a>.
   * 
   */
  //WARNING: ASSUMED T is integer
  template <typename T, typename Pr,
	    T (*onedalgoY) (int procCount, const util::Aggreg2Dto1D<T, Pr, false>& prefixSumArray, int length, int *cutIndexes, T)=oned::NicolPlus<T, util::Aggreg2Dto1D<T, Pr, false> >::nicol_plus,
	    T (*onedalgoX)(int procCount, const util::Aggreg2Dto1D<T, Pr, true>& prefixSumArray, int length, int *cutIndexes, T)=oned::NicolPlus<T, util::Aggreg2Dto1D<T, Pr, true> >::nicol_plus  >
  class JagMHeurProbeHor : public PartBase<T,Pr>
  {
    T multiarraypart(int procCount, const util::Multiarrays<T, util::Aggreg2Dto1D<T, Pr, false> > & arrs, T ub, int* procs, int** cuts);
  public:
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);

    virtual ~JagMHeurProbeHor();
  private:
  };

  /**
   * @brief Finds least max proc load value by applying binary search
   * over it. First cut direction is vertical. See section 3.2.2
   * PROBE-M algorithm of  <a href="http://arxiv.org/abs/1104.2566"> "E. Saule,
   * E. O. Bas, U. V. Catalyurek, Load-Balancing Spatially Located
   * Computations using Rectangular Partitions Paper, Tech. Report,
   * ArXiv, no. arXiv:1104.2566v1, 2011"</a>.
   * 
   */
  template <typename T, typename Pr,
	    T (*onedalgoY) (int procCount, const util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, false>& prefixSumArray, int length, int *cutIndexes, T) = oned::NicolPlus<T, util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, false> >::nicol_plus,
	    T (*onedalgoX) (int procCount, const util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>, true>& prefixSumArray, int length, int *cutIndexes, T) = oned::NicolPlus<T, util::Aggreg2Dto1D<T, util::TransposePrefix2D<T, Pr>,  true> >::nicol_plus >
  class JagMHeurProbeVer : public PartBase<T,Pr>
  {
  public: 
    virtual ~JagMHeurProbeVer();
    

    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
  };
  /**
   * @brief Finds least max proc load value by applying binary search
   * over it. First cut direction is decided by tring both horizontal
   * and vertical first dimension cuts. See section 3.2.2 PROBE-M
   * algorithm of  <a href="http://arxiv.org/abs/1104.2566"> "E. Saule,
   * E. O. Bas, U. V. Catalyurek, Load-Balancing Spatially Located
   * Computations using Rectangular Partitions Paper, Tech. Report,
   * ArXiv, no. arXiv:1104.2566v1, 2011"</a>.
   * 
   */
  template <typename T, typename Pr>
  class JagMHeurProbeBest : public PartBase<T,Pr>
  {
  public: 
    
    virtual ~JagMHeurProbeBest();

    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
  };

}

#include <twod/m_way_probe_impl.hpp>
#endif
