
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   nicol_2d.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Mon Jun 14 07:16:30 2010
 * 
 * @brief  Nicol's 2D partitioning algorithm
 * 
 * 
 */


#ifndef __NICOL_2D_H__
#define __NICOL_2D_H__

#include <util/prefix_sum_tools.hpp>
#include <util/rect_list.hpp>
#include <twod/part_base.hpp>

namespace twod
{
  /**
   * @brief Implements the rectilinear partitioning algorithm with
   * iterative refinement described in "David Nicol, Rectilinear
   * Partitioning of Irregular Data Parallel Computations, JPDC,
   * 1994".
   *
   * The number of processors in each dimension can be controlled by setP().
   * 
   * @param T type of element of the array
   * @param Pr type of the prefix sum array
   */
  template <typename T, typename Pr, bool DEBUG = false>
  class RectNicol : public PartBase<T,Pr>
  {
  private:
    static T nicol_2d_internal(int procX, int procY, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    static T fix_x_dimension(util::AggregMax2Dto1D<T,Pr,true>& array, int length, int procX, int *x_cuts);
    static T fix_y_dimension(util::AggregMax2Dto1D<T,Pr,false>& array, int length, int procY, int *y_cuts);

    ///TODO: that function should be in utils ?
    static void convertToRectList(int lengthX, int lengthY, int *x_cuts, int *y_cuts, int procX, int procY, util::RectList<T,Pr> &parts);

    int P;//number of parts in the first dimension. if 0 automatic.
  public:
    RectNicol ():P(0){}
    
    /**
     * @brief selects the number of stripes in the first dimension
     *
     * @param P number of stripes in the first dimension. If 0, the
     * number of stripes will be computed at runtime by taking the
     * square root of the number of processors. 
     */
    void setP(int P){this->P = P;}
    /** 
     * @brief Implements the rectilinear partitioning algorithm with
     * iterative refinement described in "David Nicol, Rectilinear
     * Partitioning of Irregular Data Parallel Computations, JPDC,
     * 1994"
     * 
     * @param[in] procCount number of processors.
     * @param[in] prefixSumArray input martix's prefix sum array.
     * @param[out] parts rectangle list of partition results.
     * 
     * @return Maximum loaded processor's load.
     */
    virtual T part(int procCount, const Pr& prefixSumArray, util::RectList<T,Pr> &parts);
    virtual ~RectNicol(){}
  };

}
#include <twod/nicol_2d_impl.hpp>

#endif

