
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   twod/parse.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Fri Jun 17 16:32:54 2011
 * 
 * @brief  Converts algo name string into an object
 * 
 * 
 */


#ifndef __PARSE_HPP__
#define __PARSE_HPP__
#include <string>
#include <twod/part_base.hpp>
#include "twod/rec_bisect_2d.hpp"
#include "twod/rec_bisect_relaxed.hpp"
#include "twod/uniform_2d.hpp"
#include "twod/her_jag_2d.hpp"
#include "oned/nicol_plus.hpp"
#include "twod/nicol_2d.hpp"
#include "twod/jagged_dp.hpp"
#include "twod/jagged_dp_pq.hpp"
#include "twod/m_way_jag_2d.hpp"
#include "twod/jag_pq_opt_interval.hpp"
#include "twod/hybrid.hpp"
#include "twod/m_way_probe.hpp"
#include "util/version.hpp"

namespace twod
{
  /**
   * @brief Converts text into corresponding 2D algorithm objects.
   * 
   */

  template<typename T, typename Pr>
  class Parser
  {
  public:
    /**
     * 
     * @brief Given a 2D algorithm name string, returns proper
     * partitioning algorithm object. The returned object must be
     * deallocated externally by the user using delete.
     * 
     * @param s Algorithm name. Currently valid names are: RECT-UNIFORM,
     *  RECT-NICOL, HIER-RB-LOAD, HIER-RB-DIST, HIER-RB-VER, HIER-RB-HOR
     *  HIER-RELAXED-LOAD, HIER-RELAXED-DIST, HIER-RELAXED-VER,
     *  HIER-RELAXED-HOR, HIER-RB-MIDDLE, JAG-M-OPT-HOR, JAG-M-OPT-VER,
     *  JAG-M-OPT-BEST, JAG-PQ-OPT-DP-HOR, JAG-PQ-OPT-DP-VER,
     *  JAG-PQ-OPT-DP-BEST, JAG-PQ-HEUR-VER, JAG-PQ-HEUR-HOR,
     *  JAG-PQ-HEUR-BEST, JAG-M-HEUR-PROBE-HOR, JAG-M-HEUR-PROBE-VER,
     *  JAG-M-HEUR-PROBE-BEST, JAG-PQ-OPT-INTERVAL-HOR,
     *  JAG-PQ-OPT-INTERVAL-VER, JAG-PQ-OPT-INTERVAL-BEST
     * 
     * @return Algorithm object corresponding to s
     */
    static PartBase<T,Pr>* parseName(const std::string& s);
  };
}
template<typename T, typename Pr>
twod::PartBase<T,Pr>* twod::Parser<T,Pr>::parseName(const std::string& s)
{
  twod::PartBase<T, Pr >* pb = NULL;  

  if(s.compare("NULL") == 0)
    pb = NULL;
  if(s.compare("HIER-RB-LOAD") == 0)
    pb = new twod::RecBisect2D<T, Pr, 0> ();
  else if(s.compare("HIER-RB-DIST") == 0)
    pb = new twod::RecBisect2D<T, Pr, 1> ();
  else if(s.compare("HIER-RB-VER") == 0)
    pb = new twod::RecBisect2D<T, Pr, 2> ();
  else if(s.compare("HIER-RB-HOR") == 0)
    pb = new twod::RecBisect2D<T, Pr, 3> ();
  else if(s.compare("HIER-RELAXED-LOAD") == 0)
    pb = new twod::HierRelaxed<T, Pr, 0> ();
  else if(s.compare("HIER-RELAXED-DIST") == 0)
    pb = new twod::HierRelaxed<T, Pr, 1> ();
  else if(s.compare("HIER-RELAXED-VER") == 0)
    pb = new twod::HierRelaxed<T, Pr, 2> ();
  else if(s.compare("HIER-RELAXED-HOR") == 0)
    pb = new twod::HierRelaxed<T, Pr, 3> ();
  else if(s.compare("HIER-RB-MIDDLE") == 0)
    pb = new twod::HierRelaxed<T, Pr, 1,true> ();
  else if(s.compare("RECT-UNIFORM") == 0)
    pb = new twod::Uniform<T, Pr > ();
  else if(s.compare("RECT-NICOL") == 0)
    pb = new twod::RectNicol<T, Pr > ();
  else if(s.compare("JAG-M-OPT-HOR") == 0)
    pb = new twod::JagMOptHor<T, Pr > ();
  else if(s.compare("JAG-M-OPT-VER") == 0)
    pb = new twod::JagMOptVer<T, Pr > ();
  else if(s.compare("JAG-M-OPT-BEST") == 0)
    pb = new twod::JagMOptBest<T, Pr > ();
  else if(s.compare("JAG-PQ-OPT-DP-HOR") == 0)
    pb = new twod::JagPQOptHor<T, Pr > ();
  else if(s.compare("JAG-PQ-OPT-DP-VER") == 0)
    pb = new twod::JagPQOptVer<T, Pr > ();
  else if(s.compare("JAG-PQ-OPT-DP-BEST") == 0)
    pb = new twod::JagPQOptBest<T, Pr > ();
  else if(s.compare("JAG-PQ-HEUR-VER") == 0)
    pb = new twod::JagPQHeurVer<T, Pr > ();
  else if(s.compare("JAG-PQ-HEUR-HOR") == 0)
    pb = new twod::JagPQHeurHor <T, Pr > ();
  else if(s.compare("JAG-PQ-HEUR-BEST") == 0)
    pb = new twod::JagPQHeurBest<T, Pr > ();
  else if(s.compare("JAG-M-HEUR-HOR") == 0)
    pb = new twod::JagMHeurHor<T, Pr > ();
  else if(s.compare("JAG-M-HEUR-VER") == 0)
    pb = new twod::JagMHeurVer<T, Pr > ();
  else if(s.compare("JAG-M-HEUR-BEST") == 0)
    pb = new twod::JagMHeurBest<T, Pr > ();
  else if(s.compare("JAG-M-HEUR-PROBE-HOR") == 0)
    pb = new twod::JagMHeurProbeHor<T, Pr > ();
  else if(s.compare("JAG-M-HEUR-PROBE-VER") == 0)
    pb = new twod::JagMHeurProbeVer<T, Pr > ();
  else if(s.compare("JAG-M-HEUR-PROBE-BEST") == 0)
    pb = new twod::JagMHeurProbeBest<T, Pr > ();
  else if(s.compare("JAG-PQ-OPT-INTERVAL-HOR") == 0)
    pb = new twod::JagPQOptIntervalHor<T,  Pr > ();
  else if(s.compare("JAG-PQ-OPT-INTERVAL-VER") == 0)
    pb = new twod::JagPQOptIntervalVer<T,  Pr > ();
  else if(s.compare("JAG-PQ-OPT-INTERVAL-BEST") == 0)
    pb = new twod::JagPQOptIntervalBest<T,  Pr > ();

  return pb;
}

#endif
