
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".


#ifndef __COMM_COST_2D_H__
#define __COMM_COST_2D_H__

#include "prefix_sum_tools.hpp"
#include "rect_list.hpp"

namespace util
{
/**
 * @brief Methods to measure communication cost.
 * 
 * Methods to measure communication cost which include maximum
 * neighbor, average neighbor, maximum border length and average
 * border length. A processor p2 is neighbor of processor p1 if and
 * only if p1 and p2's borders are adjacent to each other.
 */
template <typename T, typename Pr>
class Comm_cost_2d
{
private:
  static bool x_bounded(rectangle r1, rectangle r2);
  static bool y_bounded(rectangle r1, rectangle r2);
  static bool is_neighbor(rectangle r1, rectangle r2);
  static bool check_neigh(rectangle r1, rectangle r2);
public: 
  /** 
   * @brief Calculates average neighbor. 
   *
   * Total neighbor count divided by
   * processor number. The number of processors procCount and number
   * of elements in rl are independent. One rectangle can be neighbor
   * of multiple rectangles, thus counted multiple times.
   * 
   * @param rl A valid rectangle list to analyze.
   * @param procCount Number of total processors.
   * 
   * @return average neighbor value of the partition.
   */
  static double avg_neighbor(RectList<T, Pr >& rl, int procCount);
  /** 
   * @brief Finds the maximum neighbor count over all processors in rl
   * 
   * @param rl A valid rectangle list to analyze.
   * @return Maximum neighbor count.
   */
  static int max_neighbor(RectList<T, Pr>& rl);
  /** 
   * @brief Finds the total border length of all processors. 
   *
   * The outermost borders of the partition excluded.
   * 
   * @param rl A valid rectangle list to analyze.
   * @param prefixSum The prefix sum array used in building rl.
   * 
   * @return Total boerder length.
   */
  static int tot_border(RectList<T, Pr>& rl, Pr& prefixSum);
  /** 
   * @brief Finds the maximum border length over all processors.
   *
   * The outermost borders of the partition excluded.
   * 
   * @param rl A valid rectangle list to analyze.
   * @param prefixSum The prefix sum array used in building rl.
   * 
   * @return maximum border length.
   */
  static int max_border(RectList<T, Pr>& rl, Pr& prefixSum);
};
}

#include "comm_cost_2d_impl.hpp"

#endif



