
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

#ifndef PARAMETRIC_SEARCH_H__
#define PARAMETRIC_SEARCH_H__

namespace util
{
  /**
   * @brief contains several probe algorithms used for 1d partitioning.
   */

template <typename T, typename Pr>
class ParametricSearch
{
public:
  //returns 1 on success; 0 on fault
  static int probeEnhanced(T bottleneck, int cutNumber,
			   const Pr& prefixSumArray,
			   int length, int lowerIndex);

  /** 
   * Check if it is possible to partition a 1D array given processor
   * number and bottleneck.
   * 
   * @param bottleneck Bottleneck value to be tested.
   * @param cutNumber The number of cut points (That corrosponds to P-1 where P is the processor number)
   * @param prefixSumArray 1D prefix sum array with prefixSumArray[0]=0 and prefixSumArray[i]=prefixSumArray[i-1]+array[i]. array variable is the original array.
   * @param length length of prefixSumArray
   * 
   * @return 1 if it is possible to partition with given bottleneck, 0
   * if not length is the length of prefixSumArray
   */
  static int probe(T bottleneck, int cutNumber,
		   const Pr& prefixSumArray, int length);
  /** 
   * Save cut points in an array given processor number and bottleneck.
   * 
   * @param[in] bottleneck Bottleneck value to be tested.
   * @param[in] cutNumber The number of cut points (That corrosponds to P-1 where P is the processor number)
   * @param[in] prefixSumArray 1D prefix sum array with prefixSumArray[0]=0 and prefixSumArray[i]=prefixSumArray[i-1]+array[i]. array variable is the original array.
   * @param[in] length length of prefixSumArray.
   * @param[out] cuts Saved cut points. Must be preallocated to cutNumber length.
   * 
   * @return 1 if it is possible to partition with given bottleneck, 0
   * if not length is the length of prefixSumArray
   */
  static int probeCut(T bottleneck, int cutNumber,
		      const Pr& prefixSumArray, int length,
		      int *cuts);
  
  /** 
   * Check if it is possible to partition a 1D array given processor
   * number and bottleneck. described in pinar04
   * 
   * @param[in] wpre  1D prefix sum array with wpre[0]=0 and wpre[i]=wpre[i-1]+array[i]. array variable is the original array.
   * @param[in] length length of wpre
   * @param[in] bottleneck Bottleneck value to be tested.
   * @param[in] numproc Processor number.
   * @param[out] s Resulting cut points if it is possible to partition. Must be preallocated to numproc-1
   * @param[in] sl Lower bound of cut index for processor i is sl[i] for numproc and bottleneck.
   * @param[in] sh Upper bound of cut index for processor i is sl[i] for numproc and bottleneck.
   * 
   * @return 1 if it is possible to partition with given bottleneck, 0
   * if not length is the length of prefixSumArray
   */
  static int rprobe(const Pr& wpre, int length, const T& bottleneck,
		    int numproc,  int *s, const int *sl, const int *sh);
  /** 
   * Check if it is possible to partition a 1D array given processor
   * number and bottleneck. Uses interval notation defined in
   * util::PrefixSumArray1DOfTInterval.
   * 
   * @param[in] wpre  1D prefix sum array with wpre[0]=0 and wpre[i]=wpre[i-1]+array[i]. array variable is the original array.
   * @param[in] length length of wpre
   * @param[in] bottleneck Bottleneck value to be tested.
   * @param[in] numproc Processor number.
   * @param[out] s Resulting cut points if it is possible to partition. Must be preallocated to numproc-1
   * @param[in] sl Lower bound of cut index for processor i is sl[i] for numproc and bottleneck.
   * @param[in] sh Upper bound of cut index for processor i is sl[i] for numproc and bottleneck.
   * 
   * @return 1 if it is possible to partition with given bottleneck, 0
   * if not length is the length of prefixSumArray
   */
  static int rprobe_interval(const Pr& wpre, int length, const T& bottleneck,
			     int numproc, int *s, const int *sl, const int *sh);
  /** 
   * Save cut points in an array given processor number and
   * bottleneck. Uses interval notation defined in
   * util::PrefixSumArray1DOfTInterval.
   * 
   * @param[in] bottleneck Bottleneck value to be tested.
   * @param[in] cutNumber The number of cut points (That corrosponds to P-1 where P is the processor number)
   * @param[in] prefixSumArray 1D prefix sum array with prefixSumArray[0]=0 and prefixSumArray[i]=prefixSumArray[i-1]+array[i]. array variable is the original array.
   * @param[in] length length of prefixSumArray.
   * @param[out] *cuts Saved cut points. Must be preallocated to cutNumber length.
   * 
   * @return 1 if it is possible to partition with given bottleneck, 0
   * if not length is the length of prefixSumArray
   */
  static int probeCut_interval(T bottleneck, int cutNumber, const Pr& prefixSumArray, int length, int *cuts);
};
}
#include "parametric_search_impl.hpp"

#endif
