
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

#ifndef __SPART_UTIL__SUBPSA_H__
#define __SPART_UTIL__SUBPSA_H__

#include "rect_list.hpp"

namespace util
{

  /**
   * @brief represent a rectangular region of a prefix sum array as an
   * prefix sum array.
   *
   * SubPSA respect the 2D prefix sum array skeleton defined by
   * util::PrefixSumArray2DOfTBracket.
   *
   * @param T type of the data in the prefix sum array
   * @param Pr type of the prefix sum array to represent
   */
  template <typename T, typename Pr>
  class SubPSA
  {
    ///prefix sum array containing the SubPSA
    const Pr& originalpsa;
    ///region of originalpsa representend by SubPSA. Recall that rectangle are indexed from 1
    rectangle subspace;

    /**
     * @brief a row from a SubPSA.
     */
    class Intermediate
    {
      const Pr& originalpsa;
      const rectangle& subspace;
      int x;

    public:
      Intermediate (const Pr& psa, const rectangle& r, int xval)
	:originalpsa(psa), subspace(r), x(xval)
      {
      }

      T operator[] (int y) const
      {
	assert (y >= 0);
	assert (y < subspace.y_bot_r - subspace.y_top_l + 2);

	return originalpsa[x+subspace.x_top_l-1][y+subspace.y_top_l-1]
	  +originalpsa[subspace.x_top_l-1][subspace.y_top_l-1]
	  -originalpsa[x+subspace.x_top_l-1][subspace.y_top_l-1]
	  -originalpsa[subspace.x_top_l-1][y+subspace.y_top_l-1];
      }
    };

  public:
    SubPSA (const Pr& psa, const rectangle& r)
      :originalpsa(psa), subspace(r)
    {
      assert (!subspace.empty());
      assert (subspace.valid_bound(originalpsa));
    }


    /**
     * @brief size of the prefix sum array not the real matrix!
     */ 
    int prefixsizeX () const
    {
      return subspace.x_bot_r - subspace.x_top_l + 2;
    }

    /**
     * @brief size of the prefix sum array not the real matrix!
     */ 
    int prefixsizeY () const
    {
      return subspace.y_bot_r - subspace.y_top_l + 2;
    }

    Intermediate operator[] (int x) const
    {
      assert (x >= 0);
      assert (x < prefixsizeX());
      return Intermediate(originalpsa,subspace,x);
    }

    /**
     * @brief transfrom a rectangle in SubPSA space into a rectangle
     * into originalpsa space
     */
    rectangle inOriginalSpace (const rectangle& r)
    {
      //TODO: we can assert r is in subpsa (let's use a DEBUG option in the future)
      return rectangle( r.x_top_l+subspace.x_top_l-1,
			r.x_bot_r+subspace.x_top_l-1,
			r.y_top_l+subspace.y_top_l-1,
			r.y_bot_r+subspace.y_top_l-1);
      //TODO: we can assert the return value is in originalpsa

    }
  };
}

#endif
