
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

// this file is given for documentation purpose only and should not be compiled

/**
 * @brief This namespace provides multiple utilitarian classes and function
 * used in 1D and 2D partitioning algorithm (for instance as input and
 * output).
 *
 * This namespace
 * provides multiple utilitarian classes and function useful in 1D and
 * 2D partitioning algorithm. Most notably it provides implementation
 * of various PrefixSumArray for the 1D and 2D case which can be used
 * as input of the partitioning algorithm. It also provides the
 * rectangle and RectList object that constitute the output of 2D
 * partitioning algorithms.
 *
 * 1D partitioning algorithms (available in ::oned) take as an input
 * parameter a 1D prefix sum array that follows the bracket notation []
 * or the interval notation. If it follows the bracket notation, the
 * type should implement the functions described in
 * util::PrefixSumArray1DOfTBracket. If it follows the interval
 * notations, the type should implement the functions described in
 * util::PrefixSumArray1DOfTInterval.
 *
 * 2D partitioning algorithms (available in ::twod) take as an input
 * parameter a 2D prefix sum array that follows the bracket notation
 * []. A type that follows the bracket notation should implement the
 * functions described in util::PrefixSumArray2DOfTBracket. Nothing
 * prevents an algorithm to use an equivalent of the interval notation
 * in 1D in 2D. But it never appeared useful and so never was
 * implemented.
 *
 * The solution of 2D algorithms is expressed using util::RectList and
 * util::rectangle. Some communication related metrics can be computed
 * with util::Comm_cost_2d.
 *
 */
namespace util
{
  
  /**
   * @brief skeleton of prefix sum array using the bracket notation[]
   * for 1d partitioning algorithm (this class does not actually
   * exist).
   *
   * Any type that provides the following functions can be used safely
   * with 1D partitioning algorithms that use the bracket
   * notation. This class does not actually exist, it is provided for
   * documentation purpose only. Notice that since the types are given
   * as template parameter to the partitioning algorithm, there is no
   * need to inherit from this class.
   *
   * Notice that the class does not need to provide its size since it
   * is given as a parameter to the 1D partitioning functions.
   *
   * Types that respect this prototype include: double*, int*,
   * util::Aggreg2Dto1D.
   *
   * The prefix sum array of an array of primitive type (or even some
   * complex type) can be constructed using util::prefixsum.
   */
  template <typename T>
  class PrefixSumArray1DOfTBracket
  {
  public:
    /**
     * @brief returns the i-th element of a prefix sum array.
     *
     * That is to say if this is the prefix sum array of array A, we
     * have : \f$this[i] = \sum_{j=0}^{i-1} A[j]\f$. This enforces
     * that the first element is null: \f$this[0] = 0\f$ and that the
     * array is monotonically increasing: \f$this[i] \leq this[i+1]\f$.
     *
     * Notice that if the original array is of size \f$n\f$ then this
     * array should be of size \f$n+1\f$. (In a sense, it is a
     * Pascal-style array where the element at index 0 is worth 0.)
     */
    const T& operator[] (int i) const;    
  };

  /**
   * @brief skeleton of prefix sum array using the interval notation
   * for 1d partitioning algorithm (this class does not
   * actually exist).
   *
   * Any type that provides the following functions can be used safely
   * with 1D partitioning algorithms that use the interval
   * notation. This class does not actually exist, it is provided for
   * documentation purpose only. Notice that since the types are given
   * as template parameter to the partitioning algorithm, there is no
   * need to inherit from this class.
   *
   * The interval notation differs from the bracket notation since in
   * the in the bracket notation the load of an interval is equal to
   * the sum of the load of its parts. This might not be true in the
   * interval notation.
   *
   * Notice that the class does not need to provide its size since it
   * is given as a parameter to the 1D partitioning functions.
   *
   * Types that respect this prototype include:
   * util::Aggreg2Dto1Dpart, util::AggregMax2Dto1D and
   * util::Aggreg2Dto1D.
   *
   * Notice that a type that respect the bracket notation can easily
   * be extended to respect the interval notation. However, the
   * algorithm that runs on the interval notation are typically slower
   * than their bracket notation counterpart.
   */
  template <typename T>
  class PrefixSumArray1DOfTInterval
  {
  public:
    /**
     * @brief returns the load of interval \f$]i1;i2]\f$.
     *
     * i1 and i2 must be positive (or null) and i2 must be less or
     * equal to \f$n\f$ if the size of the original array is
     * \f$n\f$. In that sense, the interval notation is a
     * "Pascal-style" array. The first element of the load is given by
     * interval(0,1);
     *
     * There is no constraint that i2 must be larger than i1. For
     * instance, interval(1,0) is a valid function call that designate
     * the load of the empty interval, that is to say, that returns
     * 0. i1 can also be larger than the size of the array and
     * designate the empty interval as well.
     *
     * The only contraints on the function is that empty interval have
     * a load of 0 and the load must be increasing by inclusion; that
     * is to say if interval ]i1; i2] includes interval ]i3;i4] then
     * interval(i1,i2)>=interval(i3,i4).
     *
     * In particular, if i1<i2<i3, interval(i1,i3) might be different
     * from interval(i1,i2)+interval(i2,i3); it might be less, it
     * might be more, it might be the same.
     */
    const T& interval (int i1, int i2)const;
  };

  /**
   * @brief skeleton of prefix sum array using the bracket notation
   * for 2d partitioning algorithm (this class does not
   * actually exist).
   *
   * Any type that provides the following functions can be used safely
   * with 2D partitioning algorithms that use the bracket notation
   * (all of them right now). This class does not actually exist, it
   * is provided for documentation purpose only. Notice that since the
   * types are given as template parameter to the partitioning
   * algorithm, there is no need to inherit from this class.
   *
   * Intuitively, this class represents the prefix sum of an array (we
   * call A) so that \f$this[i][j] = \sum_{i'<i, j'<j}
   * A[i'][j']\f$. Therefore, it is similar to the "Pascal-style"; the
   * first column and the first row are composed exclusively of 0. and
   * if A if of size XxY, the prefixsum array is indexed from 0 to X
   * (included) in the first dimension and from 0 to Y (included) in
   * the second dimension.
   *
   * Types that respect this prototype include: util::Prefix2D,
   * util::SubPSA, util::TransposePrefix2D.
   *
   * One can use the function util::constraint_psa to verify the
   * validity of a 2D prefixs sum array.
   */
  template <typename T>
  class PrefixSumArray2DOfTBracket
  {
  public:
    /**
     * @brief returns the size of the first dimension of the prefix
     * sum array.
     *
     * So, if this object represents the prefix sum of an array A of
     * size XxY, this function should return X+1.
     */
    int prefixsizeX()const;
    /**
     * @brief returns the size of the second dimension of the prefix
     * sum array.
     *
     * So, if this object represents the prefix sum of an array A of
     * size XxY, this function should return Y+1.
     */
    int prefixsizeY()const;
    
    /**
     * @brief allows the access to the value of the prefix sum array.
     *
     * Since [][] is not a C++ operator, operator[] must return an
     * object whose type allow the operator[] to be applied. So that
     * one can use this[x][y]. Something must define a function of
     * prototype equivalent to "const T& operator[] (int y) const"
     * that allows value of y between 0 and prefixsizeY()-1.
     *
     * x must be between 0 and prefixsizeX()-1.
     *
     * Since the represented load matrix is of positive (or null)
     * integers, it must respect the constraints \f$\forall x' \geq x
     * \forall y' \geq y, this[x][y] \leq this[x'][y']\f$ and
     * \f$\forall x \forall y,  this[x][y] \geq 0\f$.
     */
    Something operator[] (int x) const;
  };

}
