
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// The full text of the GPL license version 2 is available in "gpl.txt".

#include "util/compact.hpp"
#include "util/prefix_sum_tools.hpp"
#include "util/rect_list.hpp"
#include <iostream>
#include <fstream>
#include <string.h>

#ifdef HAVE_BOOST
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filter/bzip2.hpp>
#endif


/** 
 * Synopsis: check_2dpart <instance> <nbproc> <cut_file>
 * instance: Begins with a line with number of tasks in X and Y dimension. The rest of the lines contain 1 task load per line. 
 * nbproc: number of processors
 * cutfile: List of rectangle cuts
 *
 * @param argc 
 * @param argv 
 * 
 * @return Given 2d instance and cut rectangles, calculates the bottleneck of the partition
 */
int main(int argc, char* argv[])
{
  using namespace util;
  if (argc != 4)
    {
      std::cerr<<"usage : <instance|-> <nbproc> <cutfile>"<<std::endl;
      return -1;
    }

  std::ifstream cutFile;
  cutFile.open(argv[3], std::ios::in);
  std::ostream& output = std::cout;
  std::ostream& err = std::cerr;
  rectangle r;
  int sizeX, sizeY;
  int procCount = atoi(argv[2]);


  typedef double T;
  typedef Prefix2D<T> Pr;

#ifdef HAVE_BOOST
  boost::iostreams::filtering_stream<boost::iostreams::input> input;
  
  std::ifstream foo ;
  
  if(strcmp(argv[1], "-") == 0)
    {
      //read from standard input
      //      input = &std::cin;
      input.push (std::cin);
    }
  else
    {
      //read from file
      
      foo.open (argv[1]);
      if (!foo.is_open())
	{
	  std::cerr<< "can not open the instance file: "<<argv[1]<<std::endl;
	  exit(-1);
	}
      {
	int len = strlen(argv[1]);
	if (len > 3 && strcmp (argv[1]+len-3, ".gz") == 0)
	  input.push(boost::iostreams::gzip_decompressor());
	if (len > 4 && strcmp (argv[1]+len-4, ".bz2") == 0)
	  input.push(boost::iostreams::bzip2_decompressor());
	
      }
      
      input.push (foo);
      //input = foo;
    }
  
  input>>sizeX>>sizeY;
  
  //reading the input
  Compact2D<T> data(sizeX, sizeY);
  for (int i=0; i<sizeX; i++)
    for (int j=0; j<sizeY; j++)
      input>>data[i][j];
  
  if (input.bad() || input.fail())
    {
      std::cerr<<"improper load matrix !"<<std::endl;
      return -1;
    }
  
  
  if(strcmp(argv[1], "-") == 0)
    {
      //      input = &std::cin;
    }
  else
    {
      //delete input;
      foo.close();
    }
  
#else
  std::istream* input;
  
  std::ifstream foo ;
  
  if(strcmp(argv[1], "-") == 0)
    {
      //read from standard input
      input = & std::cin;
    }
  else
    {
      //read from file
      
      foo.open (argv[1]);
      if (!foo.is_open())
	{
	  std::cerr<< "can not open the instance file: "<<argv[1]<<std::endl;
	  exit(-1);
	}
      
      //      input.push (foo);
      input = &foo;
    }
  
  (*input)>>sizeX>>sizeY;
  
  //reading the input
  Compact2D<T> data(sizeX, sizeY);
  for (int i=0; i<sizeX; i++)
    for (int j=0; j<sizeY; j++)
      (*input)>>data[i][j];
  
  if (input->bad() || input->fail())
    {
      std::cerr<<"improper load matrix !"<<std::endl;
      return -1;
    }

  if(strcmp(argv[1], "-") == 0)
    {
      //      input = &std::cin;
    }
  else
    {
      //delete input;
      foo.close();
    }
#endif



  Pr pr(sizeX, sizeY, data);
  RectList<T, Pr > rl(pr);
  //reading form cut file
  for(int i=0; i < procCount; i++)
    {
      cutFile >> r.x_top_l;
      cutFile >> r.y_top_l;
      cutFile >> r.x_bot_r;
      cutFile >> r.y_bot_r;
      if (cutFile.eof() || cutFile.bad() || cutFile.fail())
	break;
      rl.add_rect(r);
    }
 
  cutFile.close();

  output << rl.get_maximum_load()<<std::endl;

  if (! rl.valid_part())
    {
      err<<"partition is not valid"<<std::endl;
      return -1;
    }
  else
    return 0;
}
