
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// The full text of the GPL license version 2 is available in "gpl.txt".



#include <iostream>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include "util/tools.hpp"
#include "util/timestamp.hpp"
#include "util/prefix_sum_tools.hpp"
#include "oned/parse.hpp"

#ifdef HAVE_BOOST
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filter/bzip2.hpp>
#endif

/** 
 *Synopsis: spart_1d <algo_name> [input_array_file][-] <proc_count> 
 *algo_name: Code name of the algorithm 
 *input_array_file format: Begins with a line with number of tasks and the rest contains 1 task load per line. 
 *'-' means read from standard input
 *Notes: Edit this code to replace data type
 *
 * @param argc 
 * @param argv 
 * 
 * @return Cut indexes in one index per line format. Processor p gets all tasks from cutIndex[p-1](excluded) to cutIndex[p](included). total_load[p] = weight(cutIndex[p]) - weight(cutIndex[p-1])
 */
int main(int argc, char *argv[])
{
  using namespace oned;
  int numProc, arraySize, i;
  int *cutIndexes;

  //HEADER START
  std::cerr<<"=================================================================="<<std::endl;
  std::cerr<<"SPart, Spatially Located Workload Partitioner. Version "<<SPART_VERSION<<std::endl;
  std::cerr<<"=================================================================="<<std::endl;
  std::cerr<<"Copyright (C) 2011, The Ohio State University\ndeveloped by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek\nThis is free software with ABSOLUTELY NO WARRANTY.\nRun with no parameter for usage and licensing information."<<std::endl;
  //HEADER END


  if (argc != 4)
    {
      std::cerr<<std::endl;
      std::cerr<<"usage : "<<argv[0]<<" <algo> <inst> <nbproc>"<<std::endl;

      std::cerr<<"<algo> can be 'NICOL', 'NICOL-MINUS', 'NICOL-PLUS', 'GREEDY-BISECT' "<<std::endl;
      std::cerr<<"              'DIRECT-CUT', 'DIRECT-CUT-REF-B', 'UNIFORM'"<<std::endl;
      std::cerr<<"              'LOWER-BOUND' or 'UPPER-BOUND' (the last two have no output)."<<std::endl;
      std::cerr<<"<inst> is a filename or '-' to use standard input. It accepts .gz and .bz2 files"<<std::endl;

      std::cerr<<std::endl<<std::endl;
      std::cerr<<"LICENSING AND WARRANTY"<<std::endl<<std::endl;
      std::cerr<<"SPart is free software; you can redistribute it and/or modify it under"<<std::endl;
      std::cerr<<"the terms of the GNU General Public License as published by the Free"<<std::endl;
      std::cerr<<"Software Foundation; either version 2 or (at your option) any later"<<std::endl;
      std::cerr<<"version."<<std::endl;
      std::cerr<<""<<std::endl;
      std::cerr<<"SPart is distributed in the hope that it will be useful, but WITHOUT"<<std::endl;
      std::cerr<<"ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or"<<std::endl;
      std::cerr<<"FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License"<<std::endl;
      std::cerr<<"for more details."<<std::endl;
      std::cerr<<""<<std::endl;
      std::cerr<<"You should have received a copy of the GNU General Public License"<<std::endl;
      std::cerr<<"along with SPart; if not, write to the Free Software Foundation, Inc.,"<<std::endl;
      std::cerr<<"51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA."<<std::endl;

      return 1;
    }


  numProc = atoi(argv[3]);
  cutIndexes = new int[numProc-1];

  typedef double T;
  typedef T* Pr;

  Pr array;
  T bottleneck, max;


  T (*selected_part_func)(int, Pr const&,int, int *, T);

#ifdef HAVE_BOOST
  boost::iostreams::filtering_stream<boost::iostreams::input> input;

  std::ifstream foo; // to store the file descriptor if a file operation is needed. Allocated here because it must live past the reading operations.

  if(strcmp(argv[2], "-") == 0)
    {
      //read from standard input
      input.push(std::cin);
    }
  else
    {
      foo.open (argv[2]);
      if (!foo.is_open())
      	{
      	  std::cerr<< "can not open the instance file: "<<argv[2]<<std::endl;
      	  exit(-1);
      	}
      {
	int len = strlen(argv[2]);
	if (len > 3 && strcmp (argv[2]+len-3, ".gz") == 0)
	  input.push(boost::iostreams::gzip_decompressor());
	if (len > 4 && strcmp (argv[2]+len-4, ".bz2") == 0)
	  input.push(boost::iostreams::bzip2_decompressor());

      }

      input.push (foo);

    }

  input>>arraySize;
  
  array = new T[arraySize+1];
  array[0] = 0;
  input >> array[1];
  max = array[1];

  for(i = 2; i <= arraySize; i++)
    {
      input>>array[i];
      if (max < array[i])
	max = array[i];
      array[i] += array[i-1];
    }	

   if (input.bad() || input.fail())
    {
      std::cerr<<"improper load array !"<<std::endl;
      return -1;
    }
   
   //cleanup of input operations
   if(strcmp(argv[2], "-") == 0)
     {
       //      input = &std::cin;
     }
   else
     {
       //delete input;
       foo.close();
     }

#else
   std::istream* input;

  std::ifstream foo; // to store the file descriptor if a file operation is needed. Allocated here because it must live past the reading operations.

  if(strcmp(argv[2], "-") == 0)
    {
      //read from standard input
      input = & std::cin;
    }
  else
    {
      foo.open (argv[2]);
      if (!foo.is_open())
      	{
      	  std::cerr<< "can not open the instance file: "<<argv[2]<<std::endl;
      	  exit(-1);
      	}

      input = & foo;
    }

  (*input)>>arraySize;
  
  array = new T[arraySize+1];
  array[0] = 0;
  (*input) >> array[1];
  max = array[1];

  for(i = 2; i <= arraySize; i++)
    {
      (*input)>>array[i];
      if (max < array[i])
	max = array[i];
      array[i] += array[i-1];
    }	

   if (input->bad() || input->fail())
    {
      std::cerr<<"improper load array !"<<std::endl;
      return -1;
    }
   
   //cleanup of input operations
   if(strcmp(argv[2], "-") == 0)
     {
       //      input = &std::cin;
     }
   else
     {
       //delete input;
       foo.close();
     }

#endif

   selected_part_func = oned::Parser<T,Pr>::parseName(argv[1]);

   if (selected_part_func == NULL)
     { 
       if(strcmp(argv[1],"LOWER-BOUND") == 0)
	 {
	   std::cerr<<"time             (in seconds)        : "<< 0.0000 <<std::endl;
	   std::cerr<<"most loaded part                     : "<< array[arraySize]/numProc <<std::endl;
	   std::cerr<<"load imbalance   ((max-ideal)/ideal) : "<< 0.0000 <<std::endl;
	   return 0;
	 }
       else if(strcmp(argv[1],"UPPER-BOUND") == 0)
	 {
	   std::cerr<<"time             (in seconds)        : "<< 0.0000 <<std::endl;
	   std::cerr<<"most loaded part                     : "<< array[arraySize]/numProc+max<<std::endl;
	   std::cerr<<"load imbalance   ((max-ideal)/ideal) : "<< ( (double)max )/((double)array[arraySize]/numProc) <<std::endl;
	   return 0;
	 }
       else
	 {
	   std::cerr<<"invalid algorithm name"<<std::endl;
	   return -1;
	 }
     }

  util::timestamp t1;

  bottleneck = selected_part_func(numProc, array, arraySize + 1, cutIndexes, max);

  util::timestamp t2;

  //write cut indexes to standard output
  for(i = 0; i < numProc - 1; i++)
    std::cout<<cutIndexes[i] - 1<<std::endl;

  std::cerr<<"time             (in seconds)        : "<< t2 - t1<<std::endl;;
  std::cerr<<"most loaded part                     : "<< bottleneck<<std::endl;
  std::cerr<<"load imbalance   ((max-ideal)/ideal) : "<< (bottleneck - ((double)array[arraySize]/numProc))/((double)array[arraySize]/numProc)<<std::endl;

  
  delete[] cutIndexes;
  delete[] array;
  return 0;
}
