
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// The full text of the GPL license version 2 is available in "gpl.txt".

#include "util/compact.hpp"
#include <cstring>
#include <iostream>
#include <fstream>
#include <string>
#include <assert.h>
#include "util/tools.hpp"
#include "util/timestamp.hpp"
#include "twod/parse.hpp"
#include "twod/part_base.hpp"
#include "util/comm_cost_2d.hpp"
#include "util/rect_list.hpp"
#include <vector>

#ifdef HAVE_BOOST
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filter/bzip2.hpp>
#endif

/** 
 *Synopsis: spart_2d <algo_name> [input_array_file][-] <proc_count> 
 *algo_name: Code name of the algorithm. 
 *input_array_file format: Begins with a line with number of tasks in X and Y dimension. The rest of the lines contain 1 task load per line. 
 *'-' means read from standard input
 * @param argc 
 * @param argv 
 * 
 * @return Cut rectangles in one rectangle per line format. Processor p gets all tasks in a given partition rectangle including borders.  
 * Notes: Edit this code to replace data type
 */
int main(int argc, char* argv[])
{
  using namespace oned;
  using namespace twod;
  using namespace util;
  
  
  //HEADER START
  std::cerr<<"=================================================================="<<std::endl;
  std::cerr<<"SPart, Spatially Located Workload Partitioner. Version "<<SPART_VERSION<<std::endl;
  std::cerr<<"=================================================================="<<std::endl;
  std::cerr<<"Copyright (C) 2011, The Ohio State University\ndeveloped by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek\nThis is free software with ABSOLUTELY NO WARRANTY.\nRun with no parameter for usage and licensing information."<<std::endl;
  //HEADER END

  if (argc != 4)
    {
      std::cerr<<std::endl;
      std::cerr<<"usage : "<<argv[0]<<" <algo> <inst> <nbproc>"<<std::endl;
      std::cerr<<"<algo> can be 'RECT-NICOL', 'JAG-M-HEUR-HOR' "<<std::endl;
      std::cerr<<"<inst> is a filename or '-' to use standard input."<<std::endl;
      std::cerr<<"  It can use .gz and .bz2 files if compiled with -D HAVE_BOOST"<<std::endl;

      std::cerr<<"<algo>:"<<std::endl;
      std::cerr<<"  Rectilinear partitioning algorithms:"<<std::endl;
      std::cerr<<"  ------------------------------------"<<std::endl;
      std::cerr<<"  RECT-UNIFORM"<<std::endl;
      std::cerr<<"  RECT-NICOL"<<std::endl;
      std::cerr<<"\n  Jagged partitioning algorithms:"<<std::endl;
      std::cerr<<"  -------------------------------"<<std::endl;
      std::cerr<<"  JAG-PQ-HEUR-BEST          JAG-PQ-HEUR-HOR          JAG-PQ-HEUR-VER"<<std::endl;
      std::cerr<<"  JAG-PQ-OPT-DP-BEST        JAG-PQ-OPT-DP-HOR        JAG-PQ-OPT-DP-VER"<<std::endl;
      std::cerr<<"  JAG-PQ-OPT-INTERVAL-BEST  JAG-PQ-OPT-INTERVAL-HOR  JAG-PQ-OPT-INTERVAL-VER"<<std::endl;
      std::cerr<<"  JAG-M-HEUR-BEST           JAG-M-HEUR-HOR           JAG-M-HEUR-VER"<<std::endl;
      std::cerr<<"  JAG-M-HEUR-PROBE-BEST     JAG-M-HEUR-PROBE-HOR     JAG-M-HEUR-PROBE-VER"<<std::endl;
      std::cerr<<"  JAG-M-OPT-BEST            JAG-M-OPT-HOR            JAG-M-OPT-VER"<<std::endl;
      std::cerr<<"\n  Hierarchical partitioning algorithms:"<<std::endl;
      std::cerr<<"  -------------------------------------"<<std::endl;
      std::cerr<<"  HIER-RB-MIDDLE"<<std::endl;
      std::cerr<<"  HIER-RB-LOAD       HIER-RB-DIST       HIER-RB-HOR       HIER-RB-VER"<<std::endl;
      std::cerr<<"  HIER-RELAXED-LOAD  HIER-RELAXED-DIST  HIER-RELAXED-HOR  HIER-RELAXED-VER"<<std::endl;


      std::cerr<<std::endl;

      std::cerr<<"LICENSING AND WARRANTY"<<std::endl<<std::endl;
      std::cerr<<"SPart is free software; you can redistribute it and/or modify it under"<<std::endl;
      std::cerr<<"the terms of the GNU General Public License as published by the Free"<<std::endl;
      std::cerr<<"Software Foundation; either version 2 or (at your option) any later"<<std::endl;
      std::cerr<<"version."<<std::endl;
      std::cerr<<""<<std::endl;
      std::cerr<<"SPart is distributed in the hope that it will be useful, but WITHOUT"<<std::endl;
      std::cerr<<"ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or"<<std::endl;
      std::cerr<<"FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License"<<std::endl;
      std::cerr<<"for more details."<<std::endl;
      std::cerr<<""<<std::endl;
      std::cerr<<"You should have received a copy of the GNU General Public License"<<std::endl;
      std::cerr<<"along with SPart; if not, write to the Free Software Foundation, Inc.,"<<std::endl;
      std::cerr<<"51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA."<<std::endl;
      return -1;
    }

  int sizeX, sizeY;
  int procCount = atoi(argv[3]);

  typedef double T;
  typedef Prefix2D<T> Pr;

  twod::PartBase<T, Pr >* pb = NULL;

  std::string argv1 = argv[1];

  //   //add new developed algorithm to here

  pb = Parser<T,Pr>::parseName(argv1);

  if (pb == NULL)
    {
      if (argv1.compare("HYBRID-F--RECT-UNIFORM-JAG-M-OPT-BEST") == 0)
	{
	  twod::Hybrid<T, Pr >* h = new twod::Hybrid<T, Pr > ();
	  h->setLevel1 (Parser<T,Pr>::parseName ("RECT-UNIFORM"));
	  h->setLevel2 (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-HEUR-BEST"));
	  h->setLevel2Slow (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-OPT-BEST"));
	  pb = h;
	}
      else if (argv1.compare("HYBRID-F--RECT-NICOL-JAG-M-OPT-BEST") == 0)
	{
	  twod::Hybrid<T, Pr >* h = new twod::Hybrid<T, Pr > ();
	  h->setLevel1 (Parser<T,Pr>::parseName ("RECT-NICOL"));
	  h->setLevel2 (Parser<T,util::SubPSA<T, Pr> >::parseName("JAG-M-HEUR-BEST"));
	  h->setLevel2Slow (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-OPT-BEST"));
	  pb = h;
	}
      else if (argv1.compare("HYBRID-F--HIER-RB-LOAD-JAG-M-OPT-BEST") == 0)
	{
	  twod::Hybrid<T, Pr >* h = new twod::Hybrid<T, Pr > ();
	  h->setLevel1 (Parser<T,Pr>::parseName ("HIER-RB-LOAD"));
	  h->setLevel2 (Parser<T,util::SubPSA<T, Pr> > ::parseName("JAG-M-HEUR-BEST"));
	  h->setLevel2Slow (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-OPT-BEST"));
	  pb = h;
	}
      else if (argv1.compare("HYBRID-F--HIER-RELAXED-LOAD-JAG-M-OPT-BEST") == 0)
	{
	  twod::Hybrid<T, Pr >* h = new twod::Hybrid<T, Pr > ();
	  h->setLevel1 (Parser<T,Pr>::parseName ("HIER-RELAXED-LOAD"));
	  h->setLevel2 (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-HEUR-BEST"));
	  h->setLevel2Slow (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-OPT-BEST"));
	  pb = h;
	}
      else if (argv1.compare("HYBRID-F--JAG-PQ-HEUR-BEST-JAG-M-OPT-BEST") == 0)
	{
	  twod::Hybrid<T, Pr >* h = new twod::Hybrid<T, Pr > ();
	  h->setLevel1 (Parser<T,Pr>::parseName ("JAG-PQ-HEUR-BEST"));
	  h->setLevel2 (Parser<T,util::SubPSA<T, Pr> > ::parseName("JAG-M-HEUR-BEST"));
	  h->setLevel2Slow (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-OPT-BEST"));
	  pb = h;
	}
      else if (argv1.compare("HYBRID-F--JAG-M-HEUR-BEST-JAG-M-OPT-BEST") == 0)
	{
	  twod::Hybrid<T, Pr >* h = new twod::Hybrid<T, Pr > ();
	  h->setLevel1 (Parser<T,Pr>::parseName ("JAG-M-HEUR-BEST"));
	  h->setLevel2 (Parser<T,util::SubPSA<T, Pr> > ::parseName("JAG-M-HEUR-BEST"));
	  h->setLevel2Slow (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-OPT-BEST"));
	  pb = h;
	}
      else if (argv1.compare("HYBRID-F--JAG-M-OPT-BEST-JAG-M-OPT-BEST") == 0)
	{
	  twod::Hybrid<T, Pr >* h = new twod::Hybrid<T, Pr > ();
	  h->setLevel1 (Parser<T,Pr>::parseName ("JAG-M-OPT-BEST"));
	  h->setLevel2 (Parser<T,util::SubPSA<T, Pr> >::parseName("JAG-M-HEUR-BEST"));
	  h->setLevel2Slow (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-OPT-BEST"));
	  h->setProcLevel1(0);
	  pb = h;
	}
      else if (argv1.compare("HYBRID-F-SCAN--JAG-M-HEUR-BEST-JAG-M-OPT-BEST") == 0)
	{
	  twod::Hybrid<T, Pr >* h = new twod::Hybrid<T, Pr > ();
	  h->setLevel1 (Parser<T,Pr>::parseName ("JAG-M-HEUR-BEST"));
	  h->setLevel2 (Parser<T,util::SubPSA<T, Pr> >::parseName ("JAG-M-HEUR-BEST"));
	  h->setLevel2Slow (Parser<T,util::SubPSA<T, Pr> > ::parseName("JAG-M-OPT-BEST"));
	  h->setProcLevel1(-1);
	  pb = h;
	}
      else
	{
	  std::cerr <<"invalid algorithm name" << std::endl;
	  return -1;
	}
    }

  
#ifdef HAVE_BOOST
  boost::iostreams::filtering_stream<boost::iostreams::input> input;
  
  std::ifstream foo ;
  
  if(strcmp(argv[2], "-") == 0)
    {
      //read from standard input
      //      input = &std::cin;
      input.push (std::cin);
    }
  else
    {
      //read from file
      
      foo.open (argv[2]);
      if (!foo.is_open())
	{
	  std::cerr<< "can not open the instance file: "<<argv[2]<<std::endl;
	  exit(-1);
	}
      {
	int len = strlen(argv[2]);
	if (len > 3 && strcmp (argv[2]+len-3, ".gz") == 0)
	  input.push(boost::iostreams::gzip_decompressor());
	if (len > 4 && strcmp (argv[2]+len-4, ".bz2") == 0)
	  input.push(boost::iostreams::bzip2_decompressor());
	
      }
      
      input.push (foo);
      //input = foo;
    }
  
  input>>sizeX>>sizeY;
  
  //reading the input
  Compact2D<T> data(sizeX, sizeY);
  for (int i=0; i<sizeX; i++)
    for (int j=0; j<sizeY; j++)
      input>>data[i][j];
  
  if (input.bad() || input.fail())
    {
      std::cerr<<"improper load matrix !"<<std::endl;
      return -1;
    }
  
  
  if(strcmp(argv[2], "-") == 0)
    {
      //      input = &std::cin;
    }
  else
    {
      //delete input;
      foo.close();
    }
  
#else
  std::istream* input;
  
  std::ifstream foo ;
  
  if(strcmp(argv[2], "-") == 0)
    {
      //read from standard input
      input = & std::cin;
    }
  else
    {
      //read from file
      
      foo.open (argv[2]);
      if (!foo.is_open())
	{
	  std::cerr<< "can not open the instance file: "<<argv[2]<<std::endl;
	  exit(-1);
	}
      
      //      input.push (foo);
      input = &foo;
    }
  
  (*input)>>sizeX>>sizeY;
  
  //reading the input
  Compact2D<T> data(sizeX, sizeY);
  for (int i=0; i<sizeX; i++)
    for (int j=0; j<sizeY; j++)
      (*input)>>data[i][j];
  
  if (input->bad() || input->fail())
    {
      std::cerr<<"improper load matrix !"<<std::endl;
      return -1;
    }

  if(strcmp(argv[2], "-") == 0)
    {
      //      input = &std::cin;
    }
  else
    {
      //delete input;
      foo.close();
    }
#endif


  Pr pr(sizeX, sizeY, data);

  RectList<T, Pr > rl(pr);

  timestamp t1;

  T d = pb->part(procCount, pr, rl);

  timestamp t2;

  delete pb;

  assert (rl.valid_part());

  util::RectList<T, Pr>::container::const_iterator list_iter1 = rl.rectangles.begin();
  for(;list_iter1 != rl.rectangles.end(); list_iter1++)
    {
      const rectangle &r =  *list_iter1;
      std::cout << r.x_top_l  <<  " " << r.y_top_l << " ";
      std::cout << r.x_bot_r  <<  " " << r.y_bot_r << std::endl;
    }

  std::cerr<<"time             (in seconds)        : "<<t2-t1<<std::endl;
  std::cerr<<"most loaded part                     : "<<d<<std::endl;
  std::cerr<<"load imbalance   ((max-ideal)/ideal) : "<<((double)d-(double)(pr[sizeX][sizeY])/procCount)/(pr[sizeX][sizeY]/procCount)<<std::endl;
  //  std::cerr<<"average neighbors : "<<Comm_cost_2d<T, Pr >::avg_neighbor(rl,procCount)<<std::endl;
  //  std::cerr<<"maximum neighbors : "<<Comm_cost_2d<T, Pr >::max_neighbor(rl)<<std::endl;
  //  std::cerr<<"total border length : "<<Comm_cost_2d<T, Pr >::tot_border(rl,pr)<<std::endl;
  //  std::cerr<<"max border length : "<<Comm_cost_2d<T, Pr >::max_border(rl,pr)<<std::endl;
  return 0;
}
