
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// The full text of the GPL license version 2 is available in "gpl.txt".


#include <algorithm>
#include "util/compact.hpp"
#include "util/rect_list.hpp"
#include <iostream>
#include <sstream>
#include <fstream>
#include <cstring>
#include <assert.h>
#include <iomanip>
#ifdef HAVE_BOOST
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filter/bzip2.hpp>
#endif

#include "util/prefix_sum_tools.hpp"

double largex;
double largey;

typedef double T;
typedef util::Prefix2D<double> Pr;

using namespace util;

/** 
 * @brief Returns string in format "R,G,B" where R:Red, G:Green,
 * B:Blue is the color code.
 *
 * Returns string in format "R,G,B" where R:Red, G:Green, B:Blue is
 * the color code. This is used to draw rectangles in SVG.
 *
 * @param max maximum value in the input space. Maps to complete red color (255,0,0).
 * @param min minimum value in the input space. Maps to complete blue color (0,0,255).
 * @param val input value to be converted to color.
 * @param perfect perdect load inbalance value (that will map to white color).
 * 
 * @return Color string in "R,G,B" format.
 */
std::string getColor(T max, T min, T val, double perfect)
{
  //set this to true to make it gray scale.
  const bool gray = false;
  double intensity;
  int redScale;
  int greenScale;
  int blueScale;
 
  if(!gray)
    {
      if(val > perfect)
	{
	  intensity = (val-perfect)/(max-perfect);
	  redScale = 255;
	  greenScale = (int)((1-intensity)*255);
	  blueScale = greenScale;
	}
      else
	{
	  intensity = (val-min)/(perfect-min);
	  redScale = (int)(intensity * 255);
	  greenScale = redScale;
	  blueScale = 255;
	}
    }
  else
    {
      intensity = ((double)val-min)/((double)max-min);
      redScale = (int)(intensity*255);
      greenScale = (int)(intensity*255);
      blueScale = (int)(intensity*255); 
    }
  std::stringstream ss;

  ss<<redScale<<","<<greenScale<<","<<blueScale;

  return ss.str();
}
/** 
 * @brief Draws a single rectangle at a specified location, dimensions and color. 
 * 
 * @param x Coordinate -- distance from top.
 * @param y Coordinate -- distance from left.
 * @param w Rectangle width.
 * @param h Rectangle height.
 * @param color_in Input color string in format "R,G,B" (This is the output of getColor() function)
 * @param stroke_width Stroke width in SVG. This looks like a frame around the rectangle. (0.1 by default)
 * @param color_out Stroke color.
 */
void draw_rectangle(double x, double y, double w, double h, std::string color_in,
		    double stroke_width = 0.1, std::string color_out = "0,0,0")
{
  std::cout << "\n<rect y=\"" << x << "\" x=\"" << y << "\" width=\""<< w <<  "\" height=\"" << h
	    <<"\" style=\"fill:rgb("<< color_in <<");stroke-width:"<< stroke_width<<"; stroke:rgb("<<color_out<<")\" />";

}

/** 
 * @brief Prints text on SVG.
 * 
 * @param x Coordinate -- distance from top.
 * @param y Coordinate -- distance from left.
 * @param text Text to print.
 * @param font_size Font size in SVG.
 * @param anchor Anchor in SVG. (bottom by default)
 * @param angle To rotate text in SVG. (0 by default)
 */
void draw_text(double x, double y, std::string text, double font_size, std::string anchor="bottom", double angle=0)
{
  std::cout << "\n<text  y=\""<< x <<"\" x=\"" << y <<"\"  font-size=\"" << font_size   << "\" "
	    << "transform=\"rotate("<<angle<<" "<<y<<","<<x<<")\" "
	    <<"text-anchor=\""<<anchor<<"\" >" << text<<"</text>\n";

}

/** 
 * @brief Draws color legend to figure out load imbalance.
 * 
 * @param x Coordinate -- distance from top.
 * @param y Coordinate -- distance from left.
 * @param max Maximum value in the input space. Maps to complete red color (255,0,0).
 * @param min Minimum value in the input space. Maps to complete blue color (0,0,255). 
 * @param avg Perfect load imbalance value. Maps to complete white color (255,255,255).
 */
void draw_bar(double x, double y, T max_load, T min_load, double avg)
{
  int i=0;
  int numBarColors = 50;
  float rectHeight = x / (numBarColors + 1.3);
  float xpos;
  for(xpos = 0; i < numBarColors; xpos+=rectHeight)
    {
      i++;
      long int value = min_load + i*(max_load-min_load)/numBarColors;
      float imbalance = (value - avg)/avg;

       //in svg x represents horizontal axis and y represents vertical axis
      draw_rectangle (xpos,  (1.14)*y, largex/10, rectHeight, getColor(max_load,min_load,value,avg), 0 );
      if((i-1)%5 == 0 || i == numBarColors)
	{
	  //number corresponding to rectangle
	  std::stringstream ss;
	  ss<<value <<" (" << std::fixed <<std::setprecision(2) <<imbalance * 100 << "%)";
	  
	  draw_text (xpos + rectHeight/2, (1.25)*y , ss.str(), rectHeight, "bottom");
	}
    }
  
  draw_text (xpos + rectHeight*1.3, y*(1.14), "scale : load (imbalance)", rectHeight*1.3, "left", 0);

}

/** 
 * @brief Prints instance and cut filename, and processor count as text.
 * 
 * @param x Coordinate -- distance from top.
 * @param y Coordinate -- distance from left.
 * @param instName Instance file name.
 * @param cutName Cuts file name.
 * @param proc Processor count.
 */
void draw_footer(double x, double y, char *instName, char *cutName, int proc)
{
  std::stringstream ss;
  ss<<"Instance File:" << instName << " | Cut point file:" << cutName << " | Processor number: " << proc;

  draw_text ((1.07)*x , y, ss.str(), largex/25, "bottom");
}

/** 
 * @brief Draws a complete SVG image of Load imbalance chart with legends and other information.
 *  
 * @param rl Input rectangle list.
 * @param pr Input prefix sum array.
 * @param numProc Number of processors.
 * @param inst Instance file name.
 * @param cuts Cut file name.
 * @param min_load Minimum value in the input space. Maps to complete blue color (0,0,255). 
 * @param max_load Maximum value in the input space. Maps to complete red color (255,0,0).
 * @param legend Set to true to draw a color legend.
 */
void draw_rectangles(RectList<T, Pr >& rl,
		     const Pr& pr,
		     int numProc, char *inst, char *cuts,
		     T min_load, T max_load, bool legend )
{
  rectangle r;
  //  long int max_load = rl.get_maximum_load();
  //  long int min_load = rl.get_minimum_load();
  double average = ((double)pr[pr.prefixsizeX()-1][pr.prefixsizeY()-1]) / numProc;
  largex = 0, largey = 0;
  //std::cerr << "average:"<<average<<std::endl;
  util::RectList<T,Pr >::container::iterator list_iter1 = rl.rectangles.begin();
  for(;list_iter1 != rl.rectangles.end(); list_iter1++)
    {
       r =  *list_iter1;
       if(r.x_bot_r > largex)
	 largex = r.x_bot_r;
       if(r.y_bot_r > largey)
	 largey = r.y_bot_r;
       //in svg x represents horizontal axis and y represents vertical axis
       //std::cerr <<"rect_load: "<< r.get_load<long int,Prefix2D<long int> >(pr) << std::endl;

       std::string color = getColor(max_load,min_load,r.get_load<T,Pr >(pr), average);

       draw_rectangle (r.x_top_l, r.y_top_l,
			(r.y_bot_r - r.y_top_l) + 1, (r.x_bot_r - r.x_top_l) + 1,
			color);
    }

  if (legend)
    {
      draw_bar(largex,largey,max_load,min_load, average);
      draw_footer(largex,0,inst,cuts,numProc);
    }
}
/** 
 * @brief Prints standard SVG header to the SVG file.
 * 
 * 
 * @return SVG header that goes to the beginning of the .svg file.
 */
std::string get_svg_header()
{
  return "<?xml version=\"1.0\" standalone=\"no\"?><!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\"><svg width=\"100%\" height=\"100%%\" version=\"1.1\" xmlns=\"http://www.w3.org/2000/svg\">";
}

/** 
 * Synopsis: spart_2d_visual <instance> <nbproc> <cut_file>
 * instance: Begins with a line with number of tasks in X and Y dimension. The rest of the lines contain 1 task load per line. 
 * nbproc: number of processors
 * cutfile: List of rectangle cuts
 *
 * @param argc 
 * @param argv 
 * 
 * @return Given 2d instance and cut rectangles, outputs SVG image of load heat map or partition scheme
 */
int main(int argc, char* argv[])
{
  //HEADER START
  std::cerr<<"=================================================================="<<std::endl;
  std::cerr<<"SPart, Spatially Located Workload Partitioner. Version "<<SPART_VERSION<<std::endl;
  std::cerr<<"=================================================================="<<std::endl;
  std::cerr<<"Copyright (C) 2011, The Ohio State University\ndeveloped by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek\nThis is free software with ABSOLUTELY NO WARRANTY.\nRun with no parameter for usage and licensing information."<<std::endl;
  //HEADER END

  if (argc != 4 && argc != 6)
    {
      std::cerr<<std::endl;
      std::cerr<<"usage : <instance|-> <nbproc> <cutfile> [<min> <max>]"<<std::endl;
      std::cerr<<std::endl;
      
      std::cerr<<"LICENSING AND WARRANTY"<<std::endl<<std::endl;
      std::cerr<<"SPart is free software; you can redistribute it and/or modify it under"<<std::endl;
      std::cerr<<"the terms of the GNU General Public License as published by the Free"<<std::endl;
      std::cerr<<"Software Foundation; either version 2 or (at your option) any later"<<std::endl;
      std::cerr<<"version."<<std::endl;
      std::cerr<<""<<std::endl;
      std::cerr<<"SPart is distributed in the hope that it will be useful, but WITHOUT"<<std::endl;
      std::cerr<<"ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or"<<std::endl;
      std::cerr<<"FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License"<<std::endl;
      std::cerr<<"for more details."<<std::endl;
      std::cerr<<""<<std::endl;
      std::cerr<<"You should have received a copy of the GNU General Public License"<<std::endl;
      std::cerr<<"along with SPart; if not, write to the Free Software Foundation, Inc.,"<<std::endl;
      std::cerr<<"51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA."<<std::endl;

      return -1;
    }

  std::ifstream cutFile;
  cutFile.open(argv[3], std::ios::in);

  //  std::ifstream instFile;
  //  std::istream* instance;
  std::ostream& output = std::cout;
  std::ostream& err = std::cerr;
  rectangle r;
  int sizeX, sizeY;
  int procCount = atoi(argv[2]);






#ifdef HAVE_BOOST
  boost::iostreams::filtering_stream<boost::iostreams::input> input;
  
  std::ifstream foo ;
  
  if(strcmp(argv[1], "-") == 0)
    {
      //read from standard input
      //      input = &std::cin;
      input.push (std::cin);
    }
  else
    {
      //read from file
      
      foo.open (argv[1]);
      if (!foo.is_open())
	{
	  std::cerr<< "can not open the instance file: "<<argv[1]<<std::endl;
	  exit(-1);
	}
      {
	int len = strlen(argv[1]);
	if (len > 3 && strcmp (argv[1]+len-3, ".gz") == 0)
	  input.push(boost::iostreams::gzip_decompressor());
	if (len > 4 && strcmp (argv[1]+len-4, ".bz2") == 0)
	  input.push(boost::iostreams::bzip2_decompressor());
	
      }
      
      input.push (foo);
      //input = foo;
    }
  
  input>>sizeX>>sizeY;
  
  //reading the input
  Compact2D<T> data(sizeX, sizeY);
  for (int i=0; i<sizeX; i++)
    for (int j=0; j<sizeY; j++)
      input>>data[i][j];
  
  if (input.bad() || input.fail())
    {
      std::cerr<<"improper load matrix !"<<std::endl;
      return -1;
    }
  
  
  if(strcmp(argv[1], "-") == 0)
    {
      //      input = &std::cin;
    }
  else
    {
      //delete input;
      foo.close();
    }
  
#else
  std::istream* input;
  
  std::ifstream foo ;
  
  if(strcmp(argv[1], "-") == 0)
    {
      //read from standard input
      input = & std::cin;
    }
  else
    {
      //read from file
      
      foo.open (argv[1]);
      if (!foo.is_open())
	{
	  std::cerr<< "can not open the instance file: "<<argv[1]<<std::endl;
	  exit(-1);
	}
      
      //      input.push (foo);
      input = &foo;
    }
  
  (*input)>>sizeX>>sizeY;
  
  //reading the input
  Compact2D<T> data(sizeX, sizeY);
  for (int i=0; i<sizeX; i++)
    for (int j=0; j<sizeY; j++)
      (*input)>>data[i][j];
  
  if (input->bad() || input->fail())
    {
      std::cerr<<"improper load matrix !"<<std::endl;
      return -1;
    }

  if(strcmp(argv[1], "-") == 0)
    {
      //      input = &std::cin;
    }
  else
    {
      //delete input;
      foo.close();
    }
#endif
  


  Pr pr(sizeX, sizeY, data);
  RectList<T, Pr > rl(pr);
  //reading form cut file
  for(int i=0; i < procCount; i++)
    {
      cutFile >> r.x_top_l;
      cutFile >> r.y_top_l;
      cutFile >> r.x_bot_r;
      cutFile >> r.y_bot_r;
      if (cutFile.eof() || cutFile.bad() || cutFile.fail())
	break;
      rl.add_rect(r);
    }
 
  cutFile.close();

  std::cerr << "least loaded part                     : " << rl.get_minimum_load()<<std::endl;
  std::cerr << "most loaded part                      : " << rl.get_maximum_load()<<std::endl;
  double average = ((double)pr[pr.prefixsizeX()-1][pr.prefixsizeY()-1]) / procCount;
  std::cerr << "load imbalance    ((max-ideal)/ideal) : "<< (rl.get_maximum_load()-average)/average<<std::endl;

  if (! rl.valid_part())
    {
      err<<"partition is not valid"<<std::endl;
      return -1;
    }
  else
    {
      T max_load;
      T min_load;
      if (argc > 4)
	{
	  std::stringstream ss(argv[4]);
	  ss>>min_load;
	  std::stringstream ss2(argv[5]);
	  ss>>max_load;

	  min_load = std::min(min_load, rl.get_minimum_load());
	  max_load = std::max(max_load, rl.get_maximum_load());
	}
      else
	{
	  min_load = rl.get_minimum_load();
	  max_load = rl.get_maximum_load();
	}

      output << get_svg_header() << std::endl;
      draw_rectangles(rl,pr, procCount, argv[1], argv[3], min_load, max_load, true);
      output << "</svg>";
    }
  
}
